/*
 * Copyright (C) 2015-2016 by the University of Southern California
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <iostream>
#include <string>
#include "postman.hh"
#include "reader.hh"
#include "utility.hh"
#include "version.h"
#include <sys/wait.h>
using namespace std;

int log_level = LOG_DBG;
bool verbose_log = false;
void usage(const char *comm) {
  cerr << " Usage:\n " << comm << " OPTIONS\n"
    " -i INPUT         input, format and path separated by colon\n"
    "                  e.g. trace:PATH, text:PATH, raw:PATH\n"
    " -o OUTPUT        output, file or '-' for stdout, optional\n"
    " -a ADDRESS       listening address\n"
    " -p PORT          listening port\n"
    " -n NUMBER        number of clients to accept\n"
    " -l SECONDS       preload seconds of trace, used to control memory consumption\n"
    "                  default is none (0 or negative integer): read all in memory\n"
    " -v               verbose log\n"
    " -V               show the program version\n"
    " -h               print this message\n";
  exit(1);
}

void sig_handler (int sig)
{
  cerr << "["<< getpid()<<"]"<< ": receive signal: " << sig << endl;
}

int main(int argc, char *argv[])
{
  const char *comm = argv[0];
  if (argc == 1) usage(comm);

  string input_file, input_format, listen_addr, output_file, tmp;
  int listen_port = -1, num_clients = -1, opt = -1, status = 0, trace_limit = -1;
  int skt[2];
  pid_t child_pid, wpid, my_pid = getpid();
  size_t found;
  while ((opt = getopt(argc, argv, "i:a:p:n:o:l:vhV")) != -1) {
    switch(opt) {
    case 'i':
      tmp = optarg;
      found = tmp.find_first_of(':');
      if (found == string::npos)
	errx(1, "incorrect input format");
      input_format = tmp.substr(0, found);
      input_file = tmp.substr(found + 1);
      break;
    case 'o':
      output_file = optarg;
      break;
    case 'a':
      listen_addr = optarg;
      break;
    case 'p':
      listen_port = atoi(optarg);
      break;
    case 'l':
      trace_limit = atoi(optarg);
      break;
    case 'n':
      num_clients = atoi(optarg);
      break;
    case 'v':
      verbose_log = true;
      break;
    case 'V':
      errx(0, VERSION);
      break;
    default:
      usage(comm);
    }
  }

  if (verbose_log) {
    cout << "#trace_limit: " << trace_limit << endl;
  }

  if (input_file.length() == 0)
    errx(1, "[error] input file is invalid");
  if (input_format.length() == 0)
    errx(1, "[error] input format is invalid");
  if (input_format != "trace" && input_format != "text" && input_format != "raw")
    errx(1, "[error] input format is invalid: only 'trace','text','raw' are accepted!");
  if (listen_addr.length() == 0)
    errx(1, "[error] listening address is invalid");
  if (listen_port <= 0)
    errx(1, "[error] listening port is invalid");
  if (num_clients <= 0)
    errx(1, "[error] client number is invalid");

  LOG(LOG_DBG, "input [%s] [%s] to [%d] clients\n", input_format.c_str(), input_file.c_str(), num_clients);
  LOG(LOG_DBG, "output [%s]\n", output_file.c_str());
  LOG(LOG_DBG, "listen [%s], port [%d]\n", listen_addr.c_str(), listen_port);

  //main process *reader*: read the input file and writes to unix socket
  //child process *postman*: one event base to handle *reader* and clients' sockets
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, skt) == -1)
    err(1, "socketpair");
  LOG(LOG_DBG, "[%d] unix socket pair [%d]<->[%d]\n", my_pid, skt[0], skt[1]);
  
  child_pid = fork();
  if (child_pid < 0) {
    err(1, "[error] fork");
  } else if (child_pid == 0) {//child process: postman
    my_pid = getpid();
    LOG(LOG_DBG, "[%d] postman [%d] is up\n", my_pid, my_pid);
    Postman cmd (output_file, listen_addr, listen_port, num_clients, skt[1]);
    cmd.start();
    LOG(LOG_DBG, "[%d] ends\n", my_pid);
    return 0;
  }

  //parent process: reader
  LOG(LOG_DBG, "[%d] create postman [%d]\n", my_pid, child_pid);
  sleep(3);
  Reader rd (input_file, input_format, skt[0]);
  if (rd.unblock(POSTMAN_FLAG))
    rd.read_input(trace_limit);

  signal (SIGINT, sig_handler);
  signal (SIGTERM, sig_handler);
  LOG(LOG_DBG, "[%d] waiting for postman processes\n", my_pid);
  while ((wpid = wait(&status)) > 0) {
    LOG(LOG_DBG, "[%d] postman [%d] exits with [%d]\n", my_pid, wpid, status);
  }
  
  LOG(LOG_DBG, "[%d] ends\n", my_pid);
  return 0;
}
